from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.trials import TrialInfo, TrialValue


class Problem(ABC):
    """Problem to optimize."""

    def __init__(self, loggers: list[AbstractLogger] | None = None) -> None:
        super().__init__()

        self.loggers: list[AbstractLogger] = loggers if loggers is not None else []
        self.n_trials: float = 0
        self.n_function_calls: int = 0.0

    @property
    def f_min(self) -> float | None:
        """Return the minimum function value.

        Returns:
        -------
        float | None
            Minimum function value (if exists).
            Else, return None.
        """
        return None

    @property
    @abstractmethod
    def configspace(self) -> ConfigurationSpace:
        """Configuration Space.

        All optimizers need to receive a configspace and
        convert it to their search space definition.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        raise NotImplementedError

    @abstractmethod
    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance, name, checkpoint.

        Returns:
        -------
        TrialValue
            Value of the trial, i.e.:
                - cost : float | list[float]
                - time : float, defaults to 0.0
                - status : StatusType, defaults to StatusType.SUCCESS
                - starttime : float, defaults to 0.0
                - endtime : float, defaults to 0.0
                - additional_info : dict[str, Any], defaults to {}
        """
        raise NotImplementedError

    def evaluate(self, trial_info: TrialInfo) -> TrialValue:
        trial_value = self._evaluate(trial_info=trial_info)
        self.n_function_calls += 1
        if trial_info.normalized_budget is not None:
            self.n_trials += trial_info.normalized_budget
        else:
            self.n_trials += 1

        for logger in self.loggers:
            logger.log_trial(
                n_trials=self.n_trials,
                n_function_calls=self.n_function_calls,
                trial_info=trial_info,
                trial_value=trial_value,
            )

        return trial_value



"""* Not immeditaly obvious where I even create the "solver" for optuna
* Is it always minimize?
"""

from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import TYPE_CHECKING

import numpy as np
import optuna
from ConfigSpace import Configuration, ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    Hyperparameter,
    OrdinalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from optuna.distributions import BaseDistribution, CategoricalDistribution, FloatDistribution, IntDistribution
from optuna.samplers import TPESampler
from optuna.trial import TrialState as OptunaTrialState
from rich import print as printr

from carps.optimizers.optimizer import Optimizer
from carps.utils.pareto_front import pareto
from carps.utils.trials import StatusType, TrialInfo, TrialValue

if TYPE_CHECKING:
    from omegaconf import DictConfig
    from optuna.study import Study

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent

# NOTE: Optuna has an extra OptunaTrialState.PRUNED, which indicates something
# was halted during it's run and not progressed to the next budget. They fundamentally
# do multi-fidelity differently in that they stop currently running trials and not restart
# a trial to run until a higher budget.
optuna_trial_states: dict[StatusType, OptunaTrialState] = {
    StatusType.RUNNING: OptunaTrialState.RUNNING,
    StatusType.SUCCESS: OptunaTrialState.COMPLETE,
    StatusType.CRASHED: OptunaTrialState.FAIL,
    StatusType.TIMEOUT: OptunaTrialState.FAIL,
    StatusType.MEMORYOUT: OptunaTrialState.FAIL,
}


def hp_to_optuna_distribution(hp: Hyperparameter) -> BaseDistribution:
    """Parse a Node and its children into a ConfigurationSpace.

    Args:
        node: The Node to parse
        flat: Whether to have a hierarchical naming scheme for nodes and their children.
        conditionals: Whether to include conditionals in the space from a
            [`Choice`][amltk.pipeline.Choice]. If this is `False`, this will
            also remove all forbidden clauses and other conditional clauses.
            The primary use of this functionality is that some optimizers do not
            support these features.

            !!! TODO "Not yet supported"

                This functionality is not yet supported as we can't encode this into
                a static Optuna search space.

        delim: The delimiter to use for the names of the hyperparameters.
    """
    if isinstance(hp, UniformIntegerHyperparameter):
        return IntDistribution(hp.lower, hp.upper, log=hp.log)
    elif isinstance(hp, UniformFloatHyperparameter):
        return FloatDistribution(hp.lower, hp.upper, log=hp.log)
    elif isinstance(hp, CategoricalHyperparameter):
        if hp.weights is not None:
            raise NotImplementedError(f"Weights are not supported in Optuna ({hp})")

        return CategoricalDistribution(hp.choices)
    elif isinstance(hp, OrdinalHyperparameter):
        warnings.warn(f"Ordinal hyperparameters are not supported in Optuna, use Categorical instead for {hp}.")
        return CategoricalDistribution(hp.sequence)
    elif isinstance(hp, Constant):
        return CategoricalDistribution([hp.value])

    raise NotImplementedError(f"Can't handle hyperparameter {hp}")


class OptunaOptimizer(Optimizer):
    """An optimizer that uses Optuna to optimize a search space."""

    def __init__(
        self,
        problem: Problem,
        optuna_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)
        self._solver: Study | None = None
        self.optuna_cfg = optuna_cfg

        configspace = self.problem.configspace
        if any(configspace.forbidden_clauses):
            raise NotImplementedError("Forbidden clauses are not yet supported in Optuna")

        self.optuna_space = {hp.name: hp_to_optuna_distribution(hp) for hp in configspace.get_hyperparameters()}
        self.configspace = configspace
        self.history: dict[str, tuple[optuna.Trial, Configuration, None | float | list[float]]] = {}

    def _setup_optimizer(self) -> optuna.study.Study:
        # TODO: minimize always?
        # TODO How do I know if it's multi-objective?

        # if multiple_metrics:
        #     sampler = NSGAIISampler(seed=self.optuna_cfg.sampler.seed)
        # else
        #     sampler = TPESampler(seed=self.optuna_cfg.sampler.seed)
        """(function) def create_study(
            *,
            storage: str | BaseStorage | None = None,
            sampler: BaseSampler | None = None,
            pruner: BasePruner | None = None,
            study_name: str | None = None,
            direction: str | StudyDirection | None = None,
            load_if_exists: bool = False,
            directions: Sequence[str | StudyDirection] | None = None
        ) -> Study.
        """
        sampler = TPESampler(seed=self.optuna_cfg.sampler.seed)
        study = optuna.create_study(
            **self.optuna_cfg.study, sampler=sampler, directions=["minimize"] * self.task.n_objectives
        )
        printr(sampler)
        printr(study)

        return study

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        assert self._solver is not None

        optuna_trial: optuna.Trial = self._solver.ask(self.optuna_space)
        config = optuna_trial.params
        trial_number = optuna_trial.number
        unique_name = f"{trial_number=}"
        configspace_config = Configuration(
            configuration_space=self.configspace, values=config, allow_inactive_with_values=True
        )
        self.history[unique_name] = (optuna_trial, configspace_config, None)
        return TrialInfo(
            config=configspace_config,
            name=unique_name,
            instance=None,
            budget=None,
            seed=None,
        )

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        unique_name = trial_info.name
        assert unique_name is not None
        assert unique_name in self.history
        assert self._solver is not None

        optuna_status = optuna_trial_states[trial_value.status]

        # NOTE: Is there any case in which StatusType.RUNNING is returned, giving
        # us this OptunaTrialState.RUNNING?
        assert optuna_status is not OptunaTrialState.RUNNING

        optuna_trial, configspace_config, prev_cost = self.history[unique_name]
        assert prev_cost is None

        # Need to update the history with the cost such that we can access the cost
        # for `get_current_incumbent()`
        cost = trial_value.cost if optuna_status is OptunaTrialState.COMPLETE else None
        self.history[unique_name] = (optuna_trial, configspace_config, cost)

        self._solver.tell(trial=optuna_trial, values=cost, state=optuna_status)

    def get_pareto_front(self) -> list[tuple[TrialInfo, TrialValue]]:
        """Return the pareto front for multi-objective optimization."""
        non_none_entries = [[config, cost] for optuna_trial, config, cost in self.history.values() if cost is not None]
        costs = np.array([v[1] for v in non_none_entries])
        ids_bool = pareto(costs)
        ids = np.where(ids_bool)[0]
        return [non_none_entries[i] for i in ids]

    def get_current_incumbent(self) -> Incumbent:
        """Extract the incumbent config and cost. May only be available after a complete run.

        Returns:
        -------
        Incumbent: tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]] | None
            The incumbent configuration with associated cost.
        """
        if self.task.n_objectives == 1:
            non_none_entries = [(config, cost) for _, config, cost in self.history.values() if cost is not None]
            if len(non_none_entries) == 0:
                return None
            best_config, best_cost = min(non_none_entries, key=lambda x: x[1])
            assert not isinstance(best_cost, Iterable)
            trial_info = TrialInfo(config=best_config)
            trial_value = TrialValue(cost=best_cost)
            incumbent_tuple = (trial_info, trial_value)
        else:
            front = self.get_pareto_front()
            incumbent_tuple = [(TrialInfo(config=config), TrialValue(cost=cost)) for config, cost in front]
        return incumbent_tuple

    # NOT really needed
    def convert_configspace(self, configspace: ConfigurationSpace) -> dict[str, BaseDistribution]:
        """Convert ConfigSpace configuration space to search space from optimizer.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        SearchSpace
            Optimizer's search space.
        """
        raise NotImplementedError

    # NOT really needed
    def convert_to_trial(self, *args: tuple, **kwargs: dict) -> TrialInfo:
        """Convert proposal by optimizer to TrialInfo.

        This ensures that the problem can be evaluated with a unified API.

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        raise NotImplementedError



"""DEHB Optimizer.

* Source: https://github.com/automl/DEHB/tree/master
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from dehb import DEHB

from carps.optimizers.optimizer import Optimizer
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import Configuration, ConfigurationSpace
    from omegaconf import DictConfig

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent


class DEHBOptimizer(Optimizer):
    """An optimizer that uses DEHB to optimize an objective function."""

    def __init__(
        self,
        problem: Problem,
        dehb_cfg: DictConfig,
        task: Task,
        loggers: list[AbstractLogger] | None = None,
    ) -> None:
        super().__init__(problem, task, loggers)

        self.fidelity_enabled = True
        self.task = task
        self.dehb_cfg = dehb_cfg
        self.configspace = self.convert_configspace(problem.configspace)
        self.configspace.seed(dehb_cfg.seed)
        if self.task.max_budget is None:
            raise ValueError("max_budget must be specified to run DEHB!")
        if self.task.min_budget is None:
            raise ValueError("min_budget must be specified to run DEHB!")
        self._solver: DEHB | None = None
        self.history: dict[str, dict[str, Any]] = {}

    def _setup_optimizer(self) -> Any:
        return DEHB(
            cs=self.configspace,
            min_fidelity=self.task.min_budget,
            max_fidelity=self.task.max_budget,
            n_workers=self.task.n_workers,
            **self.dehb_cfg,
        )

    def convert_configspace(self, configspace: ConfigurationSpace) -> ConfigurationSpace:
        """Convert configuration space from Problem to Optimizer.

        Here, we don't need to convert.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        ConfigurationSpace
            Configuration space for Optimizer.
        """
        return configspace

    def convert_to_trial(  # type: ignore[override]
        self,
        config: Configuration,
        name: str | None = None,
        seed: int | None = None,
        budget: float | None = None,
    ) -> TrialInfo:
        """Convert proposal from DEHB to TrialInfo.

        This ensures that the problem can be evaluated with a unified API.

        Parameters
        ----------
        config : Configuration
            Configuration from DEHB.
        name : str, optional
            Name of the trial, by default None
        seed : int, optional
            Seed of the trial, by default None
        budget : float, optional
            Budget of the trial, by default None

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        return TrialInfo(config=config, name=name, seed=seed, budget=budget)

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        info = self.solver.ask()
        unique_name = f"{info['config_id']}_{info['fidelity']}_{self.dehb_cfg.seed}"
        self.history[unique_name] = info
        return self.convert_to_trial(
            config=info["config"],
            name=unique_name,
            seed=self.dehb_cfg.seed,
            budget=info["fidelity"],
        )

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        unique_name = trial_info.name
        assert unique_name is not None
        assert unique_name in self.history
        assert self._solver is not None

        dehb_job_info = self.history[unique_name]
        if isinstance(trial_value.cost, list):
            raise NotImplementedError("Multiobjective optimization not yet implemented for DEHB!")
        dehb_result = {"fitness": float(trial_value.cost), "cost": (trial_value.time)}
        self.solver.tell(dehb_job_info, dehb_result)

    def get_current_incumbent(self) -> Incumbent:
        """Extract the incumbent config and cost. May only be available after a complete run.

        Returns:
        -------
        Incumbent: tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]] | None
            The incumbent configuration with associated cost.
        """
        incumbent = self.solver.get_incumbents()
        inc_config = self.convert_to_trial(
            config=incumbent[0],
            seed=self.dehb_cfg.seed,
        )
        inc_value = TrialValue(cost=incumbent[1])
        return (inc_config, inc_value)



"""Implementation of HEBO Optimizer.

[2024-03-27]
Note that running `python carps/run.py +optimizer/hebo=config +problem/DUMMY=config seed=1 task.n_trials=25`
raises following error:
"linear_operator.utils.errors.NanError: cholesky_cpu: 4 of 4 elements of the torch.Size([2, 2]) tensor are NaN."

This is related to this issue: https://github.com/huawei-noah/HEBO/issues/61.

For non-dummy problems HEBO works fine.
"""

from __future__ import annotations

from collections import OrderedDict, abc
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from ConfigSpace import Configuration, ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    FloatHyperparameter,
    Hyperparameter,
    IntegerHyperparameter,
    OrdinalHyperparameter,
)
from hebo.design_space.design_space import DesignSpace
from hebo.optimizers.hebo import HEBO

from carps.optimizers.optimizer import Optimizer
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.task import Task
    from carps.utils.types import Incumbent


def configspaceHP2HEBOHP(hp: Hyperparameter) -> dict:
    """Convert ConfigSpace hyperparameter to HEBO hyperparameter.

    Parameters
    ----------
    hp : Hyperparameter
        ConfigSpace hyperparameter

    Returns:
    -------
    dict
        HEBO hyperparameter

    Raises:
    ------
    NotImplementedError
        If ConfigSpace hyperparameter is anything else than
        IntegerHyperparameter, FloatHyperparameter, CategoricalHyperparameter,
        OrdinalHyperparameter or Constant
    """
    if isinstance(hp, IntegerHyperparameter):
        if hp.log:
            return {"name": hp.name, "type": "pow_int", "lb": hp.lower, "ub": hp.upper}
        else:
            return {"name": hp.name, "type": "int", "lb": hp.lower, "ub": hp.upper}
    elif isinstance(hp, FloatHyperparameter):
        if hp.log:
            return {"name": hp.name, "type": "pow", "lb": hp.lower, "ub": hp.upper}
        else:
            return {"name": hp.name, "type": "num", "lb": hp.lower, "ub": hp.upper}
    elif isinstance(hp, CategoricalHyperparameter):
        return {"name": hp.name, "type": "cat", "categories": hp.choices}
    elif isinstance(hp, OrdinalHyperparameter):
        return {
            "name": hp.name,
            "type": "step_int",
            "lb": 0,
            "ub": len(hp.sequence),
            "step": 1,
        }
    elif isinstance(hp, Constant):
        return {"name": hp.name, "type": "cat", "categories": [hp.value]}
    else:
        raise NotImplementedError(f"Unknown hyperparameter type: {hp.__class__.__name__}")


def HEBOcfg2ConfigSpacecfg(
    hebo_suggestion: pd.DataFrame,
    design_space: DesignSpace,
    config_space: ConfigurationSpace,
    allow_inactive_with_values: bool = True,
) -> Configuration:
    """Convert HEBO config to ConfigSpace config.

    Parameters
    ----------
    hebo_suggestion : pd.DataFrame
        Configuration in HEBO format
    design_space : DesignSpace
        HEBO design space
    config_space : ConfigurationSpace
        ConfigSpace configuration space
    allow_inactive_with_values : bool
        Allow values for inactive hyperparameters. This is relevant if the space has
        conditionals but the optimizer does not support them.

    Returns:
    -------
    Configuration
        Config in ConfigSpace format

    Raises:
    ------
    ValueError
        If HEBO config is more than 1
    """
    if len(hebo_suggestion) > 1:
        raise ValueError(f"Only one suggestion is ok, got {len(hebo_suggestion)}.")
    hyp = hebo_suggestion.iloc[0].to_dict()
    for k in hyp:
        hp_type = design_space.paras[k]
        if hp_type.is_numeric and hp_type.is_discrete and not np.isnan(hyp[k]):
            hyp[k] = int(hyp[k])
            # Now we need to check if it is an ordinal hp
            hp_k = config_space.get_hyperparameter(k)
            if isinstance(hp_k, OrdinalHyperparameter):
                hyp[k] = hp_k.sequence[hyp[k]]

    return Configuration(
        configuration_space=config_space, values=hyp, allow_inactive_with_values=allow_inactive_with_values
    )


def ConfigSpacecfg2HEBOcfg(config: Configuration) -> pd.DataFrame:
    """Convert ConfigSpace config to HEBO suggestion.

    Parameters
    ----------
    config : Configuration
        Configuration

    Returns:
    -------
    pd.DataFrame
        Configuration in HEBO format, e.g.
            x1        x2
        0  2.817594  0.336420
    """
    config_dict = dict(config)
    return pd.DataFrame(config_dict, index=[0])


class HEBOOptimizer(Optimizer):
    def __init__(self, problem: Problem, task: Task, loggers: list[AbstractLogger] | None = None) -> None:
        """Interface to HEBO (https://github.com/huawei-noah/HEBO) [1].

        [1] Cowen-Rivers, Alexander I., et al. "An Empirical Study of Assumptions in Bayesian Optimisation." arXiv preprint arXiv:2012.03826 (2021).

        HEBO does not support conditional configuration spaces as well as priors for hyperparameters.


        Parameters
        ----------
        problem : Problem
            _description_
        task : Task
            The task description.

        Raises:
        ------
        ValueError
            If neither `num_trials` nor `max_wallclock_time` is specified.
        """
        super().__init__(problem, task, loggers)

        # TODO: Extend HEBO to MO (maybe just adding a config suffices)
        self.configspace = self.problem.configspace

        if len(self.configspace.get_conditions()) > 0:
            pass
            # msg = "HEBO does not support conditional search spaces."
            # raise RuntimeError(msg)

        self.hebo_configspace = self.convert_configspace(self.configspace)
        self.metric = getattr(problem, "metric", "cost")
        self.budget_type = getattr(self.problem, "budget_type", None)
        self.trial_counter = 0

        self._solver: HEBO | None = None

        self.completed_experiments: OrderedDict[int, tuple[TrialValue, TrialInfo]] = OrderedDict()

    def convert_configspace(self, configspace: ConfigurationSpace) -> DesignSpace:
        """Convert configuration space from Problem to Optimizer.

        Convert the configspace from ConfigSpace to HEBO. However, given that syne-tune does not support
        conditions and forbidden clauses, we only add hyperparameters here

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        dict[str, Any]
        -------
        DesignSpace
            HEBO design space
        """
        hps_hebo = []
        for _, v in configspace.items():
            hps_hebo.append(configspaceHP2HEBOHP(v))
        return DesignSpace().parse(hps_hebo)

    def convert_from_trial(self, trial_info: TrialInfo) -> pd.DataFrame:
        return ConfigSpacecfg2HEBOcfg(
            config=trial_info.config,
        )

    def convert_to_trial(self, rec: pd.DataFrame) -> TrialInfo:
        """Convert HEBO's recommendation to trial info.

        Parameters
        ----------
        rec : pd.DataFrame
            HEBO recommendation, can look like this for 2d:
                x1        x2
            0  2.817594  0.336420
            1 -2.293059 -1.381435
            2 -0.666595  2.016661
            3  0.130466 -3.203030

            These are four points.

        Returns:
        -------
        TrialInfo
            trial info, needed to interact with the Problem
        """
        if len(rec) > 1:
            raise ValueError(f"Only one suggestion is ok, got {len(rec)}.")
        config = HEBOcfg2ConfigSpacecfg(
            hebo_suggestion=rec,
            design_space=self.hebo_configspace,
            config_space=self.problem.configspace,
            allow_inactive_with_values=True,
        )
        return TrialInfo(config=config, instance=None, budget=None, seed=None)

    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to run.

        Returns:
        -------
        TrialInfo
            Configuration, instance, seed, budget
        """
        rec = self._solver.suggest(1)
        return self.convert_to_trial(rec=rec)

    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell: Feed experiment results back to optimizer.

        Parameters
        ----------
        trial_info : TrialInfo
            Configuration, instance, seed, budget
        trial_value : TrialValue
            Cost and additional information
        """
        cost = trial_value.cost
        suggestion = self.convert_from_trial(trial_info=trial_info)

        cost = np.asarray([cost]) if isinstance(cost, abc.Sequence) else np.asarray(cost)

        self._solver.observe(suggestion, np.asarray([cost]))

    def evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate target function.

        Store data point.

        Parameters
        ----------
        trial_info : TrialInfo
            Which config to evaluate (and instance, seed, budget)

        Returns:
        -------
        TrialValue
            Information about function evaluation
        """
        trial_value = self.problem.evaluate(trial_info=trial_info)
        self.completed_experiments[self.trial_counter] = (trial_value, trial_info)
        return trial_value

    def _setup_optimizer(self) -> HEBO:
        """Setup Optimizer.

        Returns:
        -------
        HEBO
            Instance of a HEBO Optimizer

        """
        return HEBO(space=self.hebo_configspace)

    def get_trajectory(self, sort_by: str = "trials") -> tuple[list[float], list[float]]:
        """List of x and y values of the incumbents over time. x depends on ``sort_by``.

        Parameters
        ----------
        sort_by: str
            Can be "trials" or "walltime".

        Returns:
        -------
        tuple[list[float], list[float]]

        """
        # if len(self.task.objectives) > 1:
        #     raise NotSupportedError

        X: list[int | float] = []
        Y: list[float] = []

        current_incumbent = np.inf

        for k, v in self.completed_experiments.items():
            trial_value, trial_info = v
            cost = trial_value.cost
            if cost > 1e6:
                continue
            if current_incumbent < cost:
                current_incumbent = cost

                if sort_by == "trials":
                    X.append(k)
                elif sort_by == "walltime":
                    X.append(trial_value.endtime)
                else:
                    raise RuntimeError("Unknown sort_by.")

                Y.append(cost)

        return X, Y

    def get_current_incumbent(self) -> Incumbent:
        best_x = self.solver.best_x
        best_y = self.solver.best_y
        config = HEBOcfg2ConfigSpacecfg(
            hebo_suggestion=best_x,
            design_space=self.hebo_configspace,
            config_space=self.problem.configspace,
            allow_inactive_with_values=True,
        )
        trial_info = TrialInfo(config=config)
        trial_value = TrialValue(cost=best_y)
        return (trial_info, trial_value)



from __future__ import annotations

from enum import Enum, auto
from multiprocessing import Pool
from pathlib import Path

import fire
import pandas as pd
from omegaconf import OmegaConf

from carps.analysis.gather_data import read_jsonl_content
from carps.utils.loggingutils import get_logger, setup_logging

setup_logging()
logger = get_logger(__file__)


class RunStatus(Enum):
    COMPLETED = auto()
    MISSING = auto()
    TRUNCATED = auto()


def get_experiment_status(path: Path) -> dict:
    status = RunStatus.MISSING

    cfg = OmegaConf.load(path)
    n_trials = cfg.task.n_trials
    trial_logs_fn = path.parent.parent / "trial_logs.jsonl"
    if trial_logs_fn.is_file():
        trial_logs = read_jsonl_content(str(trial_logs_fn))
        n_trials_done = trial_logs["n_trials"].max()
        status = RunStatus.COMPLETED if n_trials >= n_trials_done else RunStatus.TRUNCATED

    overrides = OmegaConf.load(path.parent / "overrides.yaml")
    # TODO maybe filter cluster
    return {
        "status": status.name,
        "benchmark_id": cfg.benchmark_id,
        "problem_id": cfg.problem_id,
        "optimizer_id": cfg.optimizer_id,
        "seed": cfg.seed,
        "overrides": " ".join(overrides),
    }


def check_missing(rundir: str, n_processes: int = 4) -> pd.DataFrame:
    rundir = Path(rundir)
    paths = rundir.glob("**/.hydra/config.yaml")
    with Pool(processes=n_processes) as pool:
        data = pool.map(get_experiment_status, paths)
    data = pd.DataFrame(data)
    data.to_csv("runstatus.csv", index=False)
    return data


def generate_commands(missing_data: pd.DataFrame, runstatus: RunStatus, rundir: str = "") -> None:
    logger.info(f"Regenerate commands for {runstatus.name} runs...")
    data = missing_data
    missing = data[data["status"].isin([runstatus.name])]
    runcommands = []
    for _gid, gdf in missing.groupby(by=["optimizer_id", "problem_id"]):
        seeds = list(gdf["seed"].unique())
        seeds.sort()
        overrides = gdf["overrides"].iloc[0].split(" ")
        overrides = [o for o in overrides if "seed" not in o]
        overrides.append(f"seed={','.join(str(s) for s in seeds)} -m")
        override = " ".join(overrides)
        runcommand = f"python -m carps.run {override}\n"
        runcommands.append(runcommand)
    runcommand_fn = Path(rundir) / f"runcommands_{runstatus.name}.sh"
    with open(runcommand_fn, "w") as file:
        file.writelines(runcommands)
    logger.info(f"Done! Regenerated runcommands at {runcommand_fn}.")


def regenerate_runcommands(rundir: str, from_cached: bool = False) -> None:
    if from_cached:
        logger.info("Loading experiment status data from 'runstatus.csv'...")
        data = pd.read_csv("runstatus.csv")
        logger.info("Done!")
    else:
        logger.info("Scanning rundirs for experiment status...")
        data = check_missing(rundir=rundir)
        logger.info("Done!")

    if len(data) > 0:
        generate_commands(data, RunStatus.MISSING, rundir)
        generate_commands(data, RunStatus.TRUNCATED, rundir)
    else:
        logger.info(f"But nothing found at {rundir}.")


if __name__ == "__main__":
    fire.Fire(regenerate_runcommands)



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from ConfigSpace import ConfigurationSpace, Float

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from carps.loggers.abstract_logger import AbstractLogger


class DummyProblem(Problem):
    def __init__(self, return_value=0, budget_type: str | None = "dummy", loggers: list[AbstractLogger] | None = None):
        super().__init__(loggers)
        self.budget_type = budget_type
        self._return_value = return_value
        self._configspace = ConfigurationSpace(
            space={
                "a": Float("a", bounds=(-1, 1)),
            }
        )

    @property
    def configspace(self) -> ConfigurationSpace:
        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        start_time = time.time()
        return TrialValue(cost=self._return_value, time=1, starttime=start_time, endtime=start_time + 1, virtual_time=1)



from __future__ import annotations

import os
from pathlib import Path

import fire
import pandas as pd

from carps.utils.index_configs import index_configs
from carps.utils.loggingutils import get_logger

logger = get_logger(__file__)

config_folder = Path(__file__).parent.parent / "configs"
config_folder_problem = config_folder / "problem"
config_folder_optimizer = config_folder / "optimizer"


def find_override(problem_id: str | None = None, optimizer_id: str | None = None):
    if problem_id is not None:
        key = "problem_id"
        path = config_folder_problem
        to_find = problem_id
    elif optimizer_id is not None:
        key = "optimizer_id"
        path = config_folder_optimizer
        to_find = optimizer_id
    else:
        raise ValueError("Please specify either `problem_id` or `optimizer_id`.")

    index_fn = path / "index.csv"
    if not index_fn.is_file():
        index_configs()
    table = pd.read_csv(index_fn)

    try:
        config_fn = table["config_fn"][table[key] == to_find].values[0]
        stripped_path = str(config_fn)[len(str(config_folder)) + 1 : -len(".yaml")]
        index = next(x for x, v in enumerate(stripped_path) if v == "/")
        return "+" + stripped_path[:index] + "=" + stripped_path[index + 1 :]
    except Exception as e:
        logger.info(f"Nothing found for {to_find} in config path {path}. Error: {e}")
        return None


def merge_overrides(overrides: list[str | None]) -> str:
    # overrides = [
    #     "+problem=YAHPO/MO/cfg_rbv2_super_1457",
    #     "+problem=YAHPO/MO/cfg_rbv2_super_1452",
    #     "+problem=YAHPO/MO/cfg_rbv2_super_1453",
    #     "+problem=YAHPO/MO/cfg_rbv2_super_1454",
    # ]
    overrides = [o for o in overrides if o is not None]
    if len(overrides) > 1:
        overrides = list(map(str, overrides))
        common = os.path.commonpath(overrides)
        index = common.find("=")
        override = common[: index + 1] + ",".join([p[index + 1 :] for p in overrides])
    else:
        override = overrides[0]
    return override


if __name__ == "__main__":
    fire.Fire(find_override)



from __future__ import annotations

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from carps.analysis.utils import get_color_palette, savefig, setup_seaborn


def plot_final_performance_boxplot(
    df: pd.DataFrame,
    x="n_trials_norm",
    y="trial_value__cost_inc_norm",
    hue="optimizer_id",
    budget_var: str = "n_trials_norm",
    max_budget: float = 1,
    figure_filename: str = "figures/final_performance_boxplot.pdf",
    figsize: tuple[int, int] = (6, 4),
    **boxplot_kwargs,
) -> tuple[plt.Figure, matplotlib.axes.Axes]:
    setup_seaborn()
    palette = get_color_palette(df)
    fig = plt.figure(figsize=figsize, dpi=300)
    ax = fig.add_subplot(111)
    ax = sns.boxplot(
        data=df[np.isclose(df[budget_var], max_budget)], y=y, x=x, hue=hue, palette=palette, ax=ax, **boxplot_kwargs
    )
    savefig(fig, figure_filename)
    return fig, ax


def plot_final_performance_violinplot(
    df: pd.DataFrame,
    x="n_trials_norm",
    y="trial_value__cost_inc_norm",
    hue="optimizer_id",
    budget_var: str = "n_trials_norm",
    max_budget: float = 1,
    figure_filename: str = "figures/final_performance_boxplot.pdf",
    figsize: tuple[int, int] = (6, 4),
    **violinplot_kwargs,
) -> tuple[plt.Figure, matplotlib.axes.Axes]:
    setup_seaborn()
    palette = get_color_palette(df)
    fig = plt.figure(figsize=figsize, dpi=300)
    ax = fig.add_subplot(111)
    ax = sns.violinplot(
        data=df[np.isclose(df[budget_var], max_budget)],
        y=y,
        x=x,
        hue=hue,
        palette=palette,
        ax=ax,
        cut=0,
        **violinplot_kwargs,
    )
    savefig(fig, figure_filename)
    return fig, ax


def create_tables(df: pd.DataFrame, budget_var: str = "n_trials_norm", max_budget: float = 1):
    perf_col_norm: str = "trial_value__cost_inc_norm"

    print(df[budget_var].max())
    df = df[np.isclose(df[budget_var], max_budget)]

    # Aggregate all

    # Calculate mean over seeds per optimizer and problem
    df_mean = df.groupby(["optimizer_id", "problem_id"])[perf_col_norm].mean()
    df_mean.name = "mean"
    df_var = df.groupby(["optimizer_id", "problem_id"])[perf_col_norm].var()
    df_var.name = "var"

    print(pd.concat((df_mean, df_var), axis=1))

    # Calculate mean over problems

    # Aggregate over benchmarks


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    import pandas as pd

    from carps.analysis.process_data import load_logs

    rundir = "/home/numina/Documents/repos/CARP-S-Experiments/lib/CARP-S/runs"

    df, df_cfg = load_logs(rundir=rundir)
    create_tables(df)



from __future__ import annotations

import json
from dataclasses import asdict
from typing import TYPE_CHECKING

from carps.loggers.abstract_logger import AbstractLogger
from carps.utils.loggingutils import get_logger, setup_logging, CustomEncoder

if TYPE_CHECKING:
    from py_experimenter.result_processor import ResultProcessor

    from carps.optimizers.optimizer import Incumbent
    from carps.utils.trials import TrialInfo, TrialValue

setup_logging()
logger = get_logger("DatabaseLogger")


def convert_trial_info(trial_info, trial_value):
    info = {"trial_info": asdict(trial_info), "trial_value": asdict(trial_value)}
    info["trial_info"]["config"] = json.dumps(asdict(trial_info)["config"].get_dictionary(), cls=CustomEncoder)
    info["trial_value"]["status"] = info["trial_value"]["status"].name
    info["trial_value"]["additional_info"] = json.dumps(info["trial_value"]["additional_info"], cls=CustomEncoder)
    info["trial_value"]["cost"] = json.dumps({"cost": json.dumps(info["trial_value"]["cost"], cls=CustomEncoder)}, cls=CustomEncoder)
    keys = ["trial_info", "trial_value"]
    for key in keys:
        d = info.pop(key)
        for k, v in d.items():
            if v is not None:
                k_new = f"{key}__{k}"
                info[k_new] = v
            else:
                # If v is None, we omit it from the dict.
                # Missing keys will automatically be filled with NULL in MySQL.
                pass
    return info


class DatabaseLogger(AbstractLogger):
    def __init__(self, result_processor: ResultProcessor | None = None) -> None:
        super().__init__()
        self.result_processor = result_processor
        if self.result_processor is None:
            logger.info("Not logging to database (result processor is None).")

    def log_trial(
        self, n_trials: int, trial_info: TrialInfo, trial_value: TrialValue, n_function_calls: int | None = None
    ) -> None:
        """Evaluate the problem and log the trial.

        Parameters
        ----------
        n_trials : float
            The number of trials that have been run so far.
            For the case of multi-fidelity, a full trial
            is a configuration evaluated on the maximum budget and
            the counter is increased by `budget/max_budget` instead
            of 1.
        trial_info : TrialInfo
            The trial info.
        trial_value : TrialValue
            The trial value.
        n_function_calls: int | None, default None
            The number of target function calls, no matter the budget.
        """
        info = convert_trial_info(trial_info, trial_value)
        info["n_trials"] = n_trials
        info["n_function_calls"] = n_function_calls if n_function_calls else n_trials

        if self.result_processor:
            self.result_processor.process_logs({"trials": info})

    def log_incumbent(self, n_trials: int, incumbent: Incumbent) -> None:
        if incumbent is None:
            return

        if not isinstance(incumbent, list):
            incumbent = [incumbent]

        for inc in incumbent:
            info = convert_trial_info(inc[0], inc[1])
            info["n_trials"] = n_trials

            if self.result_processor:
                self.result_processor.process_logs({"trajectory": info})

    def log_arbitrary(self, data: dict, entity: str) -> None:
        if self.result_processor:
            self.result_processor.process_logs({entity: data})



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import seaborn as sns

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    import pandas as pd


def get_color_palette(df: pd.DataFrame | None) -> dict[str, Any]:
    cmap = sns.color_palette("colorblind", as_cmap=True)
    optimizers = list(df["optimizer_id"].unique())
    optimizers.sort()
    return {o: cmap[i] for i, o in enumerate(optimizers)}


def savefig(fig: plt.Figure, filename: str) -> None:
    figure_filename = Path(filename)
    figure_filename.parent.mkdir(parents=True, exist_ok=True)
    fig.savefig(str(figure_filename) + ".png", dpi=300, bbox_inches="tight")
    fig.savefig(str(figure_filename) + ".pdf", dpi=300, bbox_inches="tight")


def setup_seaborn(font_scale: float | None = None) -> None:
    if font_scale is not None:
        sns.set_theme(font_scale=font_scale)
    sns.set_style("whitegrid")
    sns.set_palette("colorblind")


def filter_only_final_performance(
    df: pd.DataFrame, budget_var: str = "n_trials_norm", max_budget: float = 1, soft: bool = True
) -> pd.DataFrame:
    if not soft:
        df = df[np.isclose(df[budget_var], max_budget)]
    else:
        df = df[df.groupby(["optimizer_id", "problem_id", "seed"])[budget_var].transform(lambda x: x == x.max())]
    return df



from __future__ import annotations

from pathlib import Path

import fire
from omegaconf import OmegaConf
from py_experimenter.experiment_status import ExperimentStatus
from py_experimenter.experimenter import PyExperimenter


def main(
    pyexperimenter_configuration_file_path: str | None = None, database_credential_file_path: str | None = None
) -> None:
    experiment_configuration_file_path = (
        pyexperimenter_configuration_file_path or Path(__file__).parent.parent.parent / "container/py_experimenter.yaml"
    )

    database_credential_file_path = (
        database_credential_file_path or Path(__file__).parent.parent.parent / "container/credentials.yaml"
    )
    if database_credential_file_path is not None and not database_credential_file_path.exists():
        database_credential_file_path = None

    experimenter = PyExperimenter(
        experiment_configuration_file_path=experiment_configuration_file_path,
        name="remove_error_rows",
        database_credential_file_path=database_credential_file_path,
        log_file="logs/reset_experiments.log",
        use_ssh_tunnel=OmegaConf.load(experiment_configuration_file_path).PY_EXPERIMENTER.Database.use_ssh_tunnel,
    )
    experimenter.db_connector.reset_experiments(ExperimentStatus.ERROR.value)


if __name__ == "__main__":
    fire.Fire(main)



from __future__ import annotations

# https://github.com/Dvermetten/Many-affine-BBOB/blob/master/affine_barebones.py
import ioh
import numpy as np

scale_factors = [
    11.0,
    17.5,
    12.3,
    12.6,
    11.5,
    15.3,
    12.1,
    15.3,
    15.2,
    17.4,
    13.4,
    20.4,
    12.9,
    10.4,
    12.3,
    10.3,
    9.8,
    10.6,
    10.0,
    14.7,
    10.7,
    10.8,
    9.0,
    12.1,
]


class ManyAffine:
    def __init__(self, weights, instances, opt_loc=1, dim=5, sf_type="min_max"):
        self.weights = weights / np.sum(weights)
        self.fcts = [ioh.get_problem(fid, int(iid), dim) for fid, iid in zip(range(1, 25), instances, strict=False)]
        self.opts = [f.optimum.y for f in self.fcts]
        self.scale_factors = scale_factors
        if type(opt_loc) == int:
            self.opt_x = self.fcts[opt_loc].optimum.x
        else:
            self.opt_x = opt_loc

    def __call__(self, x):
        raw_vals = np.array(
            [
                np.clip(f(x + f.optimum.x - self.opt_x) - o, 1e-12, 1e20)
                for f, o in zip(self.fcts, self.opts, strict=False)
            ]
        )
        weighted = (np.log10(raw_vals) + 8) / self.scale_factors * self.weights
        return 10 ** (10 * np.sum(weighted) - 8)


def register_many_affine_functions():
    n_functions = 24
    dimensions = np.arange(2, 11)

    seed = 476926
    rng = np.random.default_rng(seed=seed)

    for dim in dimensions:
        for i in range(1, n_functions + 1):  # 1-based, same as original bbob
            weights = rng.uniform(size=24)
            iids = rng.integers(100, size=24)
            opt_loc = rng.uniform(size=dim) * 10 - 5  # in [-5,5] as BBOB

            f_new = ManyAffine(weights, iids, opt_loc, dim)

            name = f"MA{i}_d{dim}"

            ioh.problem.wrap_real_problem(f_new, name=name, optimization_type=ioh.OptimizationType.MIN, lb=-5, ub=5)



from __future__ import annotations

from typing import Any, Optional, Union

import numpy as np

from carps.utils.trials import TrialInfo, TrialValue

SearchSpace = Any
Cost = Union[np.ndarray, float]
Incumbent = Optional[tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]]]



from __future__ import annotations

from typing import TYPE_CHECKING

import requests
from ConfigSpace.read_and_write import json as cs_json

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from carps.loggers.abstract_logger import AbstractLogger


class ContainerizedProblemClient(Problem):
    def __init__(self, n_workers: int = 1, loggers: list[AbstractLogger] | None = None):
        super().__init__(loggers=loggers)
        self.n_workers = n_workers
        self._configspace = None

    @property
    def configspace(self) -> ConfigurationSpace:
        if self._configspace is None:
            # ask server about configspace
            response = requests.get("http://localhost:5000/configspace")
            self._configspace = cs_json.read(response.json())

        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        # ask server about evaluation
        response = requests.post("http://localhost:5000/evaluate", json=trial_info.to_json())
        return TrialValue.from_json(response.json())

    def f_min(self) -> float | None:
        raise NotImplementedError("f_min is not yet implemented for ContainerizedProblemClient")



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

from carps.analysis.utils import get_color_palette, savefig, setup_seaborn
from carps.analysis.utils import filter_only_final_performance


if TYPE_CHECKING:
    import pandas as pd

def get_order_by_mean(df: pd.DataFrame, budget_var: str = "n_trials_norm") -> list[str]:
    final_df = filter_only_final_performance(df, budget_var=budget_var)
    reduced = final_df.groupby(by="optimizer_id")["trial_value__cost_inc_norm"].apply(np.nanmean)
    reduced = reduced.sort_values()
    return reduced.index.tolist()


def plot_performance_over_time(
    df: pd.DataFrame,
    x="n_trials_norm",
    y="cost_inc_norm",
    hue="optimizer_id",
    figure_filename: str = "figures/performance_over_time.pdf",
    figsize: tuple[int, int] = (6, 4),
    show_legend: bool = True,
    title: str | None = None,
    **lineplot_kwargs,
) -> tuple[plt.Figure, matplotlib.axes.Axes]:
    setup_seaborn(font_scale=1.5)
    sorter = get_order_by_mean(df=df, budget_var=x)
    df = df.sort_values(by="optimizer_id", key=lambda column: column.map(lambda e: sorter.index(e)))
    palette = get_color_palette(df)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax = sns.lineplot(data=df, x=x, y=y, hue=hue, palette=palette, **lineplot_kwargs, ax=ax)
    if show_legend:
        ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5))
    else:
        ax.get_legend().remove()
    if title is not None:
        ax.set_title(title)
    savefig(fig, figure_filename)
    return fig, ax


def plot_rank_over_time(
    df: pd.DataFrame,
    x="n_trials_norm",
    y="cost_inc_norm",
    hue="optimizer_id",
    figure_filename: str = "figures/performance_over_time",
    figsize: tuple[int, int] = (6, 4),
    show_legend: bool = True,
    **lineplot_kwargs,
) -> tuple[plt.Figure, matplotlib.axes.Axes]:
    # TODO
    setup_seaborn(font_scale=1.5)
    sorter = get_order_by_mean(df=df)
    df = df.sort_values(by="optimizer_id", key=lambda column: column.map(lambda e: sorter.index(e)))
    palette = get_color_palette(df)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax = sns.lineplot(data=df, x=x, y=y, hue=hue, palette=palette, **lineplot_kwargs, ax=ax)
    if show_legend:
        ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5))
    else:
        ax.get_legend().remove()
    savefig(fig, figure_filename)
    return fig, ax



"""HPO-B is a benchmark for assessing the performance of black-box HPO algorithms.
Originally from HPO-B Benchmark: https://github.com/releaunifreiburg/HPO-B
Sebastian Pineda-Arango, Hadi S. Jomaa, Martin Wistuba, Josif Grabocka: HPO-B: A Large-Scale Reproducible Benchmark for
Black-Box HPO based on OpenML. NeurIPS Datasets and Benchmarks 2021
Note that this benchmark only converges the surrogate benchmark. Tabular benchmark can not be used in our framework.

To run this benchmark, you must download the benchmark surrogate model under
https://rewind.tf.uni-freiburg.de/index.php/s/rTwPgaxS2Z7NH39/download/saved-surrogates.zip
for details, please refer to https://github.com/releaunifreiburg/HPO-B
"""

from __future__ import annotations

import json
import time
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import xgboost as xgb
from ConfigSpace import ConfigurationSpace

from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from carps.loggers.abstract_logger import AbstractLogger

HPOB_SEARCH_SPACE_DIMS = {
    "151": 1,
    "5886": 9,
    "5918": 13,
    "5920": 20,
    "5923": 16,
    "5926": 10,
    "5978": 17,
    "6007": 15,
    "6073": 16,
    "6105": 53,
    "6136": 8,
    "6140": 19,
    "6154": 17,
    "6183": 41,
    "6447": 2,
    "6458": 3,
    "534": 1,
    "4796": 3,
    "5499": 12,
    "5988": 31,
    "5584": 6,
    "189": 1,
    "7064": 8,
    "5527": 8,
    "5636": 6,
    "5859": 6,
    "5860": 2,
    "5890": 2,
    "5891": 8,
    "5906": 16,
    "5965": 10,
    "5968": 8,
    "5970": 2,
    "5971": 16,
    "6308": 2,
    "6762": 6,
    "6766": 2,
    "6767": 18,
    "6794": 10,
    "6856": 1,
    "7188": 1,
    "7189": 3,
    "7190": 3,
    "7607": 9,
    "7609": 9,
    "5624": 7,
    "124": 1,
    "153": 1,
    "243": 1,
    "245": 1,
    "246": 1,
    "247": 1,
    "248": 1,
    "423": 1,
    "506": 1,
    "633": 1,
    "2553": 1,
    "4006": 1,
    "5526": 2,
    "7286": 2,
    "7290": 4,
    "5626": 17,
    "5889": 6,
    "5458": 16,
    "5489": 14,
    "4828": 10,
    "214": 1,
    "6741": 21,
    "158": 1,
    "2566": 2,
    "3894": 6,
    "6003": 18,
    "6765": 2,
    "5963": 15,
    "6000": 2,
    "6322": 8,
    "5623": 3,
    "5969": 8,
    "3737": 2,
    "829": 4,
    "833": 5,
    "935": 4,
    "6323": 2,
    "7200": 19,
    "673": 1,
    "674": 1,
    "678": 1,
    "679": 1,
    "680": 1,
    "681": 1,
    "682": 1,
    "683": 1,
    "684": 1,
    "685": 1,
    "688": 1,
    "689": 1,
    "690": 1,
    "691": 1,
    "692": 1,
    "693": 1,
    "694": 1,
    "695": 1,
    "696": 1,
    "697": 1,
    "3994": 1,
    "5964": 9,
    "5972": 2,
    "6075": 6,
    "2010": 8,
    "2039": 14,
    "2073": 13,
    "2277": 15,
    "3489": 12,
    "3490": 29,
    "3502": 6,
    "3960": 15,
    "4289": 20,
    "5218": 19,
    "5237": 8,
    "5253": 2,
    "5295": 13,
    "5301": 15,
    "5315": 9,
    "2614": 2,
    "2629": 6,
    "2793": 12,
    "2799": 7,
    "2823": 11,
    "3414": 2,
    "3425": 9,
    "3434": 19,
    "3439": 3,
    "3442": 5,
    "5503": 2,
    "6131": 2,
    "5435": 3,
    "7021": 3,
    "5502": 8,
    "5521": 3,
    "5604": 12,
    "5704": 3,
    "5788": 3,
    "5813": 10,
    "7680": 2,
    "7604": 2,
    "5919": 5,
    "5921": 3,
    "5922": 10,
    "5960": 7,
    "6024": 10,
    "6124": 18,
    "6134": 11,
    "6137": 18,
    "6139": 16,
    "6155": 7,
    "6156": 1,
    "6182": 3,
    "6189": 20,
    "6190": 6,
    "6211": 15,
    "6212": 4,
    "6213": 3,
    "6215": 10,
    "6216": 5,
    "6271": 18,
    "6285": 10,
    "6309": 7,
    "6345": 20,
    "6347": 2,
    "6364": 3,
    "6365": 4,
    "6376": 7,
    "6433": 18,
    "6461": 8,
    "6493": 15,
    "6507": 3,
}


class HPOBProblem(Problem):
    def __init__(
        self,
        dataset_id: tuple[str, int],
        model_id: tuple[str, int],
        surrogates_dir: Path = Path("carps/benchmark_data/HPO-B/saved-surrogates"),
        loggers: list[AbstractLogger] | None = None,
    ):
        """Constructor for the HPO-B handler. Given that the configuration space of HPO-B tabular dataset is not generated
        from grid, we only consider surrogate benchmark.
        Parameters.
        ----------
            dataset_id: tuple[str, int]
                dataset id, the ids can be found under surrogate_model summary directory
            model_id: tuple[str, int]
                model id, each model corresponds to a search space, the search space dimensions can be found under
            surrogates_dir: Path
                path to directory with surrogates models.
        """
        super().__init__(loggers)
        self.model_id = str(model_id)
        self.dataset_id = str(dataset_id)

        surrogates_dir = Path(surrogates_dir)
        surrogates_file = surrogates_dir / "summary-stats.json"
        if not surrogates_file.is_file():
            raise RuntimeError(
                "It seems that the surrogate files have not been downloaded. Please run "
                "'bash container_recipes/benchmarks/HPOB/download_data.sh' to download the "
                "surrogates."
            )
        with open(str(surrogates_file)) as f:
            self.surrogates_stats = json.load(f)
        self.surrogate_dir = surrogates_dir

        self.surrogate_model = self._get_surrogate_model(self.dataset_id, self.model_id)

        # generate configuration space, all the feature values range from [0,1] according to the setting
        search_space_dims = HPOB_SEARCH_SPACE_DIMS[model_id]
        self.search_space_dims = search_space_dims

        self._configspace = self._get_configspace(search_space_dims)

    def _get_surrogate_model(self, dataset_id: str, model_id: str) -> xgb.Booster:
        """Get the surrogate model for the problem."""
        surrogate_name = "surrogate-" + str(model_id) + "-" + str(dataset_id)
        surrogate_dir = self.surrogate_dir / (surrogate_name + ".json")
        if not surrogate_dir.exists():
            raise ValueError(f"Unknown dataset: {dataset_id} and model: {model_id} combination")
        bst_surrogate = xgb.Booster()
        bst_surrogate.load_model(str(self.surrogate_dir / (surrogate_name + ".json")))
        return bst_surrogate

    def _get_configspace(self, search_space_dims: int):
        # generate configuration space, all the feature values range 0 to 1
        bounds = tuple([(0, 1) for _ in range(search_space_dims)])
        cs = ConfigurationSpace()
        cs.generate_all_continuous_from_bounds(bounds)
        return cs

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        TrialValue
            Cost
        """
        configuration = trial_info.config
        input = np.asarray(list(dict(configuration).values()))
        starttime = time.time()
        x_q = xgb.DMatrix(input.reshape(-1, self.search_space_dims))
        predicted_output = self.surrogate_model.predict(x_q)
        endtime = time.time()
        T = endtime - starttime

        # we would like to do minimization
        return TrialValue(cost=-predicted_output.item(), time=T, starttime=starttime, endtime=endtime)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ConfigSpace import CategoricalHyperparameter, ConfigurationSpace, Constant, OrdinalHyperparameter
from ConfigSpace.hyperparameters import (
    BetaFloatHyperparameter,
    BetaIntegerHyperparameter,
    NormalFloatHyperparameter,
    NormalIntegerHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from dataclasses_json import dataclass_json


def get_search_space_info(configspace: ConfigurationSpace) -> dict[str, Any]:
    hps = dict(configspace)
    dimension = len(hps)
    search_space_has_priors = False
    n_categoricals = 0
    n_integers = 0
    n_ordinals = 0
    n_floats = 0
    for hp in hps.values():
        if isinstance(hp, CategoricalHyperparameter):
            n_categoricals += 1
        elif isinstance(hp, BetaIntegerHyperparameter | NormalIntegerHyperparameter | UniformIntegerHyperparameter):
            n_integers += 1
        elif isinstance(hp, BetaFloatHyperparameter | NormalFloatHyperparameter | UniformFloatHyperparameter):
            n_floats += 1
        elif isinstance(hp, OrdinalHyperparameter):
            n_ordinals += 1
        elif isinstance(hp, Constant):
            dimension -= 1

        if isinstance(
            hp,
            BetaFloatHyperparameter
            | BetaIntegerHyperparameter
            | NormalFloatHyperparameter
            | NormalIntegerHyperparameter,
        ):
            search_space_has_priors = True

    assert n_categoricals + n_floats + n_integers + n_ordinals == dimension

    search_space_has_conditionals = len(configspace.get_conditions()) > 0
    search_space_has_forbiddens = len(configspace.get_forbiddens()) > 0
    return {
        "dimensions": dimension,
        "search_space_n_categoricals": n_categoricals,
        "search_space_n_ordinals": n_ordinals,
        "search_space_n_integers": n_integers,
        "search_space_n_floats": n_floats,
        "search_space_has_conditionals": search_space_has_conditionals,
        "search_space_has_forbiddens": search_space_has_forbiddens,
        "search_space_has_priors": search_space_has_priors,
    }


@dataclass_json
@dataclass(frozen=True)
class Task:
    """Task information.

    For general optimization, only `n_trials` or `time_budget` needs
    to be defined. The optimizers receive the search space.
    For multi-fidelity, at least `is_multifidelity` and `max_budget´
    need to be specified.
    For multi-objecitve, at least `n_objectives` needs to be specified.
    The remaining parameters are meta-data and not necessarily needed
    by the optimizer but useful to order tasks.

    Parameters
    ----------
    # General
    n_trials : int
        The number of trials aka calls to the objective function.
        Specify this for classic blackbox problems.
        Either `n_trials´ or `time_budget` needs to be specified.
    time_budget : float
        The time budget in minutes for optimization.
        Specify this for multi-fidelity problems.
        Either `n_trials´ or `time_budget` needs to be specified.

    # Parallelism
    n_workers : int = 1
        The number of workers allowed for this task. Not every optimizer
        allows parallelism.

    # Multi-objective
    n_objectives : int
        The number of optimization objectives.
    objectives : list[str]
        Optional names of objectives.

    # Multi-fidelity
    is_multifidelity : bool
        Whether the task is a multi-fidelity problem.
    fidelity_type : str
        The kind of fidelity used.
    min_budget : float
        Minimum fidelity. Not used by every optimizer.
    max_budget : float
        Maximum fidelity. Required for multi-fidelity.

    has_constraints : bool
        Whether the task has any constraints.

    # Objective Function Characteristics
    domain : str
        The task's domain, e.g. synethetic, ML, NAS.
    objective_function_approximation : str
        How the objective function is approximated / represented, e.g.
        real, surrogate or tabular.
    has_virtual_time : bool
        Whether the task tracked evaluation time in the case of surrogate
        or tabular objective functions.
    deterministic : bool
        Whether the objective function is deterministic.

    # Search Space Information
    dimensions: int
        The dimensionality of the task.
    search_space_n_categoricals: int
        The number of categorical hyperparameters (HPs).
    search_space_n_ordinals: int
        The number of ordinal HPs.
    search_space_n_integers: int
        The number of integer HPs.
    search_space_n_floats: int
        The number of float HPs.
    search_space_has_conditionals: bool
        Whether the search space has conditions. Not every optimizer
        supports conditional search spaces.
    search_space_has_forbiddens: bool
        Whether the search space has forbiddens/constraints.
        Not every optimizer supports forbiddens.
    search_space_has_priors: bool
        Whether there are any priors on HPs, e.g. beta or normal.

    Raises:
    ------
    ValueError
        When `is_multifidelity` is set and `max_budget` not specified.
        In order to use multi-fidelity, both need to be specified.
    ValueError
        When neither `n_trials` nor `time_budget` are specified.
    """

    # General (REQUIRED)
    n_trials: int | None = None
    time_budget: float | None = None  # 1 cpu, walltime budget in minutes

    # Parallelism
    n_workers: int = 1

    # Multi-Objective
    n_objectives: int | None = None
    objectives: list[str] | None = None

    # Multi-Fidelity
    is_multifidelity: bool | None = None
    fidelity_type: str | None = None
    min_budget: float | None = None
    max_budget: float | None = None

    # Constraint BO
    has_constraints: bool | None = None

    # Objective Function Characteristics
    domain: str | None = None  # e.g. synthetic, ML, NAS, x
    objective_function_approximation: str | None = None  # real, surrogate, tabular
    has_virtual_time: bool | None = None
    deterministic: bool | None = None

    # Search Space
    dimensions: int | None = None
    search_space_n_categoricals: int | None = None
    search_space_n_ordinals: int | None = None
    search_space_n_integers: int | None = None
    search_space_n_floats: int | None = None
    search_space_has_conditionals: bool | None = None
    search_space_has_forbiddens: bool | None = None
    search_space_has_priors: bool | None = None

    def __post_init__(self):
        if self.is_multifidelity and self.max_budget is None:
            raise ValueError("Please specify max budget for multifidelity.")
        if self.n_trials is None and self.time_budget is None:
            raise ValueError("Please specify either `n_trials` or `time_budget`.")



from __future__ import annotations

import json
import logging
import os
import traceback
import warnings
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING

from hydra.core.hydra_config import HydraConfig
from hydra.types import RunMode

from carps.loggers.abstract_logger import AbstractLogger
from carps.utils.loggingutils import get_logger, setup_logging, CustomEncoder

if TYPE_CHECKING:
    from carps.optimizers.optimizer import Incumbent
    from carps.utils.trials import TrialInfo, TrialValue

setup_logging()
logger = get_logger("FileLogger")


def get_run_directory() -> Path:
    """Get current run dir.

    Either '.' if hydra not active, else hydra run dir.

    Returns:
    -------
    Path
        Directory
    """
    try:
        # Check if we are in a hydra context
        hydra_cfg = HydraConfig.instance().get()
        if hydra_cfg.mode == RunMode.RUN:
            return Path(hydra_cfg.run.dir)

        # TODO: How can we check to actually make sure it's a multi-run...
        # MULTIRUN
        return Path(hydra_cfg.sweep.dir) / hydra_cfg.sweep.subdir
    except Exception as e:
        cwd = Path.cwd()
        tb = traceback.format_exc()
        msg = f"Unexpected issue getting current run_directory!\n{tb}\n{e}\n\nReturning current directory of {cwd}."
        warnings.warn(msg, category=UserWarning, stacklevel=2)
        return cwd


def dump_logs(log_data: dict, filename: str, directory: str | Path | None = None):
    """Dump log dict in jsonl format.

    This appends one json dict line to the filename.

    Parameters
    ----------
    log_data : dict
        Data to log, must be json serializable.
    filename : str
        Filename without path. The path will be either the
        current working directory or if it is called during
        a hydra session, the hydra run dir will be the log
        dir.
    directory: str | None, defaults to None
        Directory to log to. If None, either use hydra run dir or current dir.
    """
    log_data_str = json.dumps(log_data, cls=CustomEncoder) + "\n"
    _dir = Path(directory) if directory is not None else get_run_directory()
    filepath = _dir / filename
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with filepath.open("a") as file:
        file.writelines([log_data_str])


def convert_trials(n_trials, trial_info, trial_value, n_function_calls: int | None = None):
    if n_function_calls is None:
        n_function_calls = n_trials
    info = {
        "n_trials": n_trials,
        "n_function_calls": n_function_calls,
        "trial_info": asdict(trial_info),
        "trial_value": asdict(trial_value),
    }
    info["trial_info"]["config"] = list(dict(info["trial_info"]["config"]).values())
    info["trial_value"]["virtual_time"] = float(info["trial_value"]["virtual_time"])
    return info


class FileLogger(AbstractLogger):
    _filename: str = "trial_logs.jsonl"
    _filename_trajectory: str = "trajectory_logs.jsonl"

    def __init__(self, overwrite: bool = False, directory: str | Path | None = None) -> None:
        """File logger.

        For each trial/function evaluate, write one line to the file.
        This line contains the json serialized info dict with `n_trials`,
        `trial_info` and `trial_value`.

        Parameters
        ------
        overwrite: bool, defaults to True
            Delete previous logs in that directory if True.
            If false, raise an error message.
        directory: str | None, defaults to None
            Directory to log to. If None, either use hydra run dir or current dir.

        """
        super().__init__()

        directory = Path(directory) if directory is not None else get_run_directory()
        self.directory = directory
        if (directory / self._filename).is_file():
            if overwrite:
                logger.info(f"Found previous run. Removing '{directory}'.")
                for root, _dirs, files in os.walk(directory):
                    for f in files:
                        full_fn = os.path.join(root, f)
                        if ".hydra" not in full_fn:
                            os.remove(full_fn)
                            logger.debug(f"Removed {full_fn}")
            else:
                raise RuntimeError(
                    f"Found previous run at '{directory}'. Stopping run. If you want to overwrite, specify overwrite "
                    f"for the file logger in the config (CARP-S/carps/configs/logger.yaml)."
                )

    def log_trial(
        self, n_trials: int, trial_info: TrialInfo, trial_value: TrialValue, n_function_calls: int | None = None
    ) -> None:
        """Evaluate the problem and log the trial.

        Parameters
        ----------
        n_trials : float
            The number of trials that have been run so far.
            For the case of multi-fidelity, a full trial
            is a configuration evaluated on the maximum budget and
            the counter is increased by `budget/max_budget` instead
            of 1.
        trial_info : TrialInfo
            The trial info.
        trial_value : TrialValue
            The trial value.
        n_function_calls: int | None, default None
            The number of target function calls, no matter the budget.
        """
        info = convert_trials(n_trials, trial_info, trial_value, n_function_calls)
        if logger.level >= logging.DEBUG:
            info_str = json.dumps(info, cls=CustomEncoder) + "\n"
            logger.debug(info_str)
        else:
            info_str = f"n_trials: {info['n_trials']}, config: {info['trial_info']['config']}, cost: {info['trial_value']['cost']}"
            if info["trial_info"]["budget"] is not None:
                info_str += f" budget: {info['trial_info']['budget']}"
            logger.info(info_str)

        dump_logs(log_data=info, filename=self._filename, directory=self.directory)

    def log_incumbent(self, n_trials: int, incumbent: Incumbent) -> None:
        if incumbent is None:
            return
        if not isinstance(incumbent, list):
            incumbent = [incumbent]

        for inc in incumbent:
            info = convert_trials(n_trials, inc[0], inc[1])
            dump_logs(log_data=info, filename=self._filename_trajectory, directory=self.directory)

    def log_arbitrary(self, data: dict, entity: str) -> None:
        dump_logs(log_data=data, filename=f"{entity}.jsonl", directory=self.directory)



"""This module is a hook that when any code is being rendered, it will
print the path to the file being rendered.

This makes it easier to identify which file is being rendered when an error happens.
"""
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

import mkdocs
import mkdocs.plugins

if TYPE_CHECKING:
    import mkdocs.structure.pages

log = logging.getLogger("mkdocs")

RENDER_EXAMPLES_ENV_VAR = "AMLTK_DOC_RENDER_EXAMPLES"
EXEC_DOCS_ENV_VAR = "AMLTK_EXEC_DOCS"

truthy_values = {"yes", "on", "true", "1", "all"}


def on_pre_page(
    page: mkdocs.structure.pages.Page,
    config: Any,
    files: Any,
) -> mkdocs.structure.pages.Page | None:
    render_examples = os.environ.get(RENDER_EXAMPLES_ENV_VAR, "true")
    render_code = os.environ.get(EXEC_DOCS_ENV_VAR, "true")
    if render_examples.lower() in truthy_values or render_code.lower() in truthy_values:
        log.info(f"{page.file.src_path}")



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from omegaconf import DictConfig, OmegaConf

from carps.utils.task import Task
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from ConfigSpace import ConfigurationSpace

    from carps.benchmarks.problem import Problem
    from carps.loggers.abstract_logger import AbstractLogger
    from carps.utils.types import Incumbent, SearchSpace


class Optimizer(ABC):
    def __init__(
        self, problem: Problem, task: Task | dict | DictConfig, loggers: list[AbstractLogger] | None = None
    ) -> None:
        super().__init__()
        self.problem = problem

        if isinstance(task, dict):
            task = Task(**task)
        elif isinstance(task, DictConfig):
            task = Task(**OmegaConf.to_container(cfg=task, resolve=True))
        elif isinstance(task, Task):
            pass
        else:
            raise ValueError("task must be either `Task`, `dict` or `DictConfig`.")

        self.task: Task = task
        self.loggers: list[AbstractLogger] = loggers if loggers is not None else []

        # Convert min to seconds
        self.time_budget = self.task.time_budget * 60 if self.task.time_budget is not None else None
        self.virtual_time_elapsed_seconds: float | None = 0.0
        self.trial_counter: float = 0

        # This indicates if the optimizer can deal with multi-fidelity optimization
        self.fidelity_enabled = False

        self._solver: Any = None
        self._last_incumbent: tuple[TrialInfo, TrialValue] | None = None

    @property
    def solver(self) -> Any:
        return self._solver

    @solver.setter
    def solver(self, value: Any) -> None:
        self._solver = value

    def setup_optimizer(self):
        self.solver = self._setup_optimizer()

    @abstractmethod
    def _setup_optimizer(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def convert_configspace(self, configspace: ConfigurationSpace) -> SearchSpace:
        """Convert ConfigSpace configuration space to search space from optimizer.

        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space from Problem.

        Returns:
        -------
        SearchSpace
            Optimizer's search space.
        """
        raise NotImplementedError

    @abstractmethod
    def convert_to_trial(self, *args: tuple, **kwargs: dict) -> TrialInfo:
        """Convert proposal by optimizer to TrialInfo.

        This ensures that the problem can be evaluated with a unified API.

        Returns:
        -------
        TrialInfo
            Trial info containing configuration, budget, seed, instance.
        """
        raise NotImplementedError

    @abstractmethod
    def get_current_incumbent(self) -> Incumbent:
        """Extract the incumbent config and cost. May only be available after a complete run.

        Returns:
        -------
        Incumbent: tuple[TrialInfo, TrialValue] | list[tuple[TrialInfo, TrialValue]] | None
            The incumbent configuration with associated cost.
        """
        raise NotImplementedError

    def run(self) -> Incumbent:
        if self.solver is None:
            self.setup_optimizer()
        return self._run()

    def _time_left(self, start_time) -> bool:
        return (time.time() - start_time) + self.virtual_time_elapsed_seconds < self.time_budget

    def continue_optimization(self, start_time) -> bool:
        cont = True
        if self.time_budget is not None and not self._time_left(start_time):
            cont = False
        if self.task.n_trials is not None and self.trial_counter >= self.task.n_trials:
            cont = False

        return cont

    def _run(self) -> Incumbent:
        """Run Optimizer on Problem."""
        start_time = time.time()
        while self.continue_optimization(start_time=start_time):
            trial_info = self.ask()
            normalized_budget = 1
            if self.task.max_budget is not None and trial_info.budget is not None:
                normalized_budget = trial_info.budget / self.task.max_budget
            if self.task.is_multifidelity:
                trial_info = TrialInfo(
                    config=trial_info.config,
                    instance=trial_info.instance,
                    seed=trial_info.seed,
                    budget=trial_info.budget,
                    normalized_budget=normalized_budget,
                    checkpoint=trial_info.checkpoint,
                    name=trial_info.name,
                )
            trial_value = self.problem.evaluate(trial_info=trial_info)
            self.virtual_time_elapsed_seconds += trial_value.virtual_time
            self.tell(trial_info=trial_info, trial_value=trial_value)

            new_incumbent = self.get_current_incumbent()
            if new_incumbent != self._last_incumbent:
                self._last_incumbent = new_incumbent
                for logger in self.loggers:
                    logger.log_incumbent(self.trial_counter, new_incumbent)

            if not self.task.is_multifidelity:
                self.trial_counter += 1
            else:
                self.trial_counter += trial_info.budget / self.task.max_budget
        return self.get_current_incumbent()

    @abstractmethod
    def ask(self) -> TrialInfo:
        """Ask the optimizer for a new trial to evaluate.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Returns:
        -------
        TrialInfo
            trial info (config, seed, instance, budget)
        """
        raise NotImplementedError

    @abstractmethod
    def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
        """Tell the optimizer a new trial.

        If the optimizer does not support ask and tell,
        raise `carps.utils.exceptions.AskAndTellNotSupportedError`
        in child class.

        Parameters
        ----------
        trial_info : TrialInfo
            trial info (config, seed, instance, budget)
        trial_value : TrialValue
            trial value (cost, time, ...)
        """
        raise NotImplementedError

    # def convert_configspace(self, configspace: ConfigurationSpace) -> SearchSpace:
    # def convert_to_trial(self, *args: tuple, **kwargs: dict) -> TrialInfo:
    # def ask(self) -> TrialInfo:
    # def tell(self, trial_info: TrialInfo, trial_value: TrialValue) -> None:
    # def get_current_incumbent(self) -> Incumbent:



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import TYPE_CHECKING

import fire
from omegaconf import OmegaConf
from py_experimenter.experiment_status import ExperimentStatus
from py_experimenter.experimenter import PyExperimenter

from carps.utils.loggingutils import get_logger, setup_logging
from carps.utils.requirements import check_requirements
from carps.utils.running import optimize

if TYPE_CHECKING:
    from py_experimenter.result_processor import ResultProcessor

setup_logging()
logger = get_logger("Run from DB")


def py_experimenter_evaluate(parameters: dict, result_processor: ResultProcessor, custom_config: dict):
    try:
        config = parameters["config"]
        cfg_dict = json.loads(config)

        job_id = getattr(os.environ, "SLURM_JOB_ID", None)
        if job_id is not None:
            result_processor.process_results({"slurm_job_id": job_id})

        cfg = OmegaConf.create(cfg_dict)

        check_requirements(cfg=cfg)

        # os.chdir(cfg.outdir)

        optimize(cfg, result_processor=result_processor)

        status = ExperimentStatus.DONE
    except Exception as e:
        print(e)
        status = ExperimentStatus.ERROR
        raise

    return status


def main(
    pyexperimenter_configuration_file_path: str | None = None, database_credential_file_path: str | None = None
) -> None:
    slurm_job_id = getattr(os.environ, "SLURM_JOB_ID", None)

    experiment_configuration_file_path = (
        pyexperimenter_configuration_file_path or Path(__file__).parent / "container/py_experimenter.yaml"
    )

    database_credential_file_path = (
        database_credential_file_path or Path(__file__).parent / "container/credentials.yaml"
    )
    if database_credential_file_path is not None and not database_credential_file_path.exists():
        database_credential_file_path = None

    experimenter = PyExperimenter(
        experiment_configuration_file_path=experiment_configuration_file_path,
        name="example_notebook",
        database_credential_file_path=database_credential_file_path,
        log_file=f"logs/{slurm_job_id}.log",
        use_ssh_tunnel=OmegaConf.load(experiment_configuration_file_path).PY_EXPERIMENTER.Database.use_ssh_tunnel,
    )

    experimenter.execute(py_experimenter_evaluate, max_experiments=1)


if __name__ == "__main__":
    fire.Fire(main)



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any

import ioh
from ConfigSpace import ConfigurationSpace, Float

from carps.benchmarks.manyaffinebbob import register_many_affine_functions
from carps.benchmarks.problem import Problem
from carps.utils.trials import TrialInfo, TrialValue

if TYPE_CHECKING:
    from carps.loggers.abstract_logger import AbstractLogger

register_many_affine_functions()


class BBOBProblem(Problem):
    def __init__(self, fid: int, instance: int, dimension: int, seed: int, loggers: list[AbstractLogger] | None = None):
        super().__init__(loggers)

        self._configspace, self._problem = get_bbob_problem(fid=fid, instance=instance, dimension=dimension, seed=seed)

    @property
    def f_min(self) -> float | None:
        return self._problem.optimum.y

    @property
    def configspace(self) -> ConfigurationSpace:
        """Return configuration space.

        Returns:
        -------
        ConfigurationSpace
            Configuration space.
        """
        return self._configspace

    def _evaluate(self, trial_info: TrialInfo) -> TrialValue:
        """Evaluate problem.

        Parameters
        ----------
        trial_info : TrialInfo
            Dataclass with configuration, seed, budget, instance.

        Returns:
        -------
        float
            Cost
        """
        configuration = trial_info.config
        input = list(dict(configuration).values())
        starttime = time.time()
        cost = self._problem(input)
        endtime = time.time()
        T = endtime - starttime

        return TrialValue(cost=cost, time=T, starttime=starttime, endtime=endtime)


def get_bbob_problem(fid: int, instance: int, dimension: int, seed: int) -> tuple[ConfigurationSpace, Any]:
    r"""Get BBOB problem.

    Parameters
    ----------
    fid : int
        Function id $\in$ [1,24].
    instance : int
        Function instance.
    dimension : int
        Dimension. 1-x
    seed : int
        Seed for configuration space.

    Returns:
    -------
    tuple[ConfigurationSpace, Any]
        Configuration space, target function.
    """
    problem = ioh.get_problem(
        fid=fid,
        instance=instance,
        dimension=dimension,
        # problem_type=ProblemType.BBOB,
    )

    # Configuration space
    lower_bounds = problem.bounds.lb
    upper_bounds = problem.bounds.ub
    n_dim = problem.meta_data.n_variables
    hps = [Float(name=f"x{i}", bounds=[lower_bounds[i], upper_bounds[i]]) for i in range(n_dim)]
    configuration_space = ConfigurationSpace(seed=seed)
    configuration_space.add_hyperparameters(hps)

    return configuration_space, problem



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import matplotlib
import numpy as np
import pandas as pd
import seaborn as sns
from hydra.core.utils import setup_globals
from rich.logging import RichHandler
import matplotlib.pyplot as plt


def setup_logging() -> None:
    FORMAT = "%(message)s"
    logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])

def get_logger(logger_name: str) -> logging.Logger:
    """Get the logger by name."""
    return logging.getLogger(logger_name)

setup_logging()
logger = get_logger(__file__)


setup_globals()

def normalize(S: pd.Series, epsilon: float = 1e-8) -> pd.Series:
    return (S - S.min()) / (S.max() - S.min() + epsilon)

def convert_mixed_types_to_str(logs: pd.DataFrame, logger=None) -> pd.DataFrame:
    mixed_type_columns = logs.select_dtypes(include=["O"]).columns
    if logger:
        logger.debug(f"Goodybe all mixed data, ruthlessly converting {mixed_type_columns} to str...")
    for c in mixed_type_columns:
        # D = logs[c]
        # logs.drop(columns=c)
        if c == "cfg_str":
            continue
        logs[c] = logs[c].map(lambda x: str(x))
        logs[c] = logs[c].astype("str")
    return logs


def normalize_logs(logs: pd.DataFrame) -> pd.DataFrame:
    logger.info("Start normalization...")
    logger.info("Normalize n_trials...")
    logs["n_trials_norm"] = logs.groupby("problem_id")["n_trials"].transform(normalize)
    logger.info("Normalize cost...")
    # Handle MO
    ids_mo = logs["scenario"]=="multi-objective"
    if len(ids_mo) > 0 and "hypervolume" in logs:
        hv = logs.loc[ids_mo, "hypervolume"]
        logs.loc[ids_mo, "trial_value__cost"] = -hv  # higher is better
        logs["trial_value__cost"] = logs["trial_value__cost"].astype("float64")
        logs["trial_value__cost_inc"] = logs["trial_value__cost"].transform("cummin")
    logs["trial_value__cost_norm"] = logs.groupby("problem_id")["trial_value__cost"].transform(normalize)
    logger.info("Calc normalized incumbent cost...")
    logs["trial_value__cost_inc_norm"] = logs.groupby(by=["problem_id", "optimizer_id", "seed"])["trial_value__cost_norm"].transform("cummin")
    if "time" not in logs:
        logs["time"] = 0
    logger.info("Normalize time...")
    logs["time_norm"] = logs.groupby("problem_id")["time"].transform(normalize)
    logs = convert_mixed_types_to_str(logs, logger)
    logger.info("Done.")
    return logs

def get_interpolated_performance_df(logs: pd.DataFrame, n_points: int = 20, x_column: str = "n_trials_norm", interpolation_columns: list[str] | None = None) -> pd.DataFrame:
    """Get performance dataframe for plotting.

    Interpolated at regular intervals.

    Parameters
    ----------
    logs : pd.DataFrame
        Preprocessed logs.
    n_points : int, optional
        Number of interpolation steps, by default 20
    x_column : str, optional
        The x-axis column to interpolate by, by default 'n_trials_norm'

    Raises:
    ------
    ValueError
        When x_column missing in dataframe.

    Returns:
    -------
    pd.DataFrame
        Performance data frame for plotting
    """
    if interpolation_columns is None:
        interpolation_columns = ["trial_value__cost", "trial_value__cost_norm", "trial_value__cost_inc", "trial_value__cost_inc_norm"]
    logger.info("Create dataframe for neat plotting by aligning x-axis / interpolating budget.")

    if x_column not in logs:
        msg = f"x_column `{x_column}` not in logs! Did you call `carps.analysis.process_data.process_logs` on the raw logs?"
        raise ValueError(msg)

    # interpolation_columns = [
    #     c for c in logs.columns if c != x_column and c not in identifier_columns and not c.startswith("problem")]
    group_keys = ["scenario", "set", "benchmark_id", "optimizer_id", "problem_id", "seed"]
    x = np.linspace(0, 1, n_points + 1)
    D = []
    for gid, gdf in logs.groupby(by=group_keys):
        metadata = dict(zip(group_keys, gid, strict=False))
        performance_data = {}
        performance_data[x_column] = x
        for icol in interpolation_columns:
            if icol in gdf:
                xp = gdf[x_column].to_numpy()
                fp = gdf[icol].to_numpy()
                y = np.interp(x=x, xp=xp, fp=fp)
                performance_data[icol] = y
        performance_data.update(metadata)
        D.append(pd.DataFrame(performance_data))
    return pd.concat(D).reset_index(drop=True)





def get_color_palette(df: pd.DataFrame | None) -> dict[str, Any]:
    cmap = sns.color_palette("colorblind", as_cmap=True)
    optimizers = list(df["optimizer_id"].unique())
    optimizers.sort()
    return {o: cmap[i] for i,o in enumerate(optimizers)}

def savefig(fig: plt.Figure, filename: str) -> None:
    figure_filename = Path(filename)
    figure_filename.parent.mkdir(parents=True, exist_ok=True)
    fig.savefig(str(figure_filename) + ".png", dpi=300, bbox_inches="tight")
    fig.savefig(str(figure_filename) + ".pdf", dpi=300, bbox_inches="tight")

def setup_seaborn(font_scale: float | None = None) -> None:
    if font_scale is not None:
        sns.set_theme(font_scale=font_scale)
    sns.set_style("whitegrid")
    sns.set_palette("colorblind")

def filter_only_final_performance(df: pd.DataFrame, budget_var: str = "n_trials_norm", max_budget: float = 1, soft: bool = True) -> pd.DataFrame:
    if not soft:
        df = df[np.isclose(df[budget_var], max_budget)]
    else:
        df = df[df.groupby(["optimizer_id", "problem_id", "seed"])[budget_var].transform(lambda x: x == x.max())]
    return df


def plot_performance_over_time(df: pd.DataFrame, x="n_trials_norm", y="cost_inc_norm", hue="optimizer_id", figure_filename: str = "figures/performance_over_time.pdf", figsize: tuple[int,int]=(6,4), show_legend: bool = True, **lineplot_kwargs
                               ) -> tuple[plt.Figure, matplotlib.axes.Axes]:
    setup_seaborn(font_scale=1.5)
    palette = get_color_palette(df)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax = sns.lineplot(data=df, x=x, y=y, hue=hue, palette=palette, **lineplot_kwargs, ax=ax)
    if show_legend:
        ax.legend(loc="center left", bbox_to_anchor=(1.05, 0.5))
    else:
        ax.get_legend().remove()
    savefig(fig, figure_filename)
    return fig, ax

def get_df_crit(df: pd.DataFrame, budget_var: str = "n_trials_norm", max_budget: float = 1, soft: bool = True, perf_col: str = "trial_value__cost_inc", remove_nan: bool = True) -> pd.DataFrame:
    df = filter_only_final_performance(df=df, budget_var=budget_var, max_budget=max_budget, soft=soft)

    # Work on mean of different seeds
    df_crit = df.groupby(["optimizer_id", "problem_id"])[perf_col].apply(np.nanmean).reset_index()

    df_crit = df_crit.pivot(
        index="problem_id",
        columns="optimizer_id",
        values=perf_col
    )

    if remove_nan:
        lost = df_crit[np.array([np.any(np.isnan(d.values)) for _, d in df_crit.iterrows()])]

        # Rows are problems, cols are optimizers
        df_crit = df_crit[np.array([not np.any(np.isnan(d.values)) for _, d in df_crit.iterrows()])]
        logger.info(f"Lost following experiments: {lost}")

    return df_crit


if __name__ == "__main__":
    # df.columns:
    # Index(['n_trials', 'n_incumbents', 'trial_value__cost',
    #        'trial_value__cost_inc', 'scenario', 'benchmark_id', 'problem_id',
    #        'optimizer_id', 'seed', 'reference_point', 'hypervolume', 'set'],
    #       dtype='object')
    # all but following necessary (but you can adjust also what the function works on): n_incumbents, scenario, reference_point, hypervolume
    df = pd.read_parquet("runs_subset_BB/dev/logs.parquet")
    df["set"] = "dev"

    df = normalize_logs(df)
    # print_overview(df)
    perf = get_interpolated_performance_df(df)
    perf_time = get_interpolated_performance_df(df, x_column="time_norm")



    lineplot_kwargs = {"linewidth": 3}
    for gid, gdf in perf.groupby(by=["scenario", "set"]):
        print(gid)
        fig, ax = plot_performance_over_time(
            df=gdf,
            x="n_trials_norm",
            y="trial_value__cost_inc_norm",
            hue="optimizer_id",
            figure_filename=f"figures/perf_over_time/performance_over_time_{gid}_trials.pdf",
            figsize=(6,4),
            **lineplot_kwargs
        )

        perf_col = "trial_value__cost_inc_norm"
        problem_prefix = ""
        fpath = Path("figures")
        fpath.mkdir(exist_ok=True, parents=True)
        identifier = gid

        # DF on normalized perf values
        df_crit = get_df_crit(gdf, remove_nan=False, perf_col=perf_col)
        # df_crit = df_crit.reindex(columns=names) # NOTE: Can order columns here
        df_crit.index = [i.replace(problem_prefix + "/dev/", "") for i in df_crit.index]
        df_crit.index = [i.replace(problem_prefix + "/test/", "") for i in df_crit.index]
        plt.figure(figsize=(12, 12))
        sns.heatmap(df_crit, annot=False, fmt="g", cmap="viridis_r")
        plt.title("Performance of Optimizers per Problem (Normalized)")
        plt.ylabel("Problem ID")
        plt.xlabel("Optimizer")
        savefig(plt.gcf(), fpath / f"perf_opt_per_problem_{identifier}")
        plt.show()

        # Df on raw values
        # Optionally, plot the ranked data as a heatmap
        df_crit = get_df_crit(gdf, remove_nan=False, perf_col=perf_col)
        # df_crit = df_crit.reindex(columns=names) # NOTE: Can order columns here
        df_crit.index = [i.replace(problem_prefix + "/dev/", "") for i in df_crit.index]
        df_crit.index = [i.replace(problem_prefix + "/test/", "") for i in df_crit.index]
        ranked_df = df_crit.rank(axis=1, method="min", ascending=True)

        plt.figure(figsize=(12, 12))
        sns.heatmap(ranked_df, annot=True, fmt="g", cmap="viridis_r")
        plt.title("Ranking of Optimizers per Problem")
        plt.ylabel("Problem ID")
        plt.xlabel("Optimizer")
        savefig(plt.gcf(), fpath / f"rank_opt_per_problem_{identifier}")
        plt.show()

        # Plotting the heatmap of the rank correlation matrix
        correlation_matrix = ranked_df.corr(method="spearman")
        plt.figure(figsize=(8,6))
        sns.heatmap(correlation_matrix, annot=True, cmap="coolwarm", cbar=True, square=True, fmt=".2f")
        plt.title("Spearman Rank Correlation Matrix Between Optimizers")
        savefig(plt.gcf(), fpath / f"spearman_rank_corr_matrix_opt_{identifier}")
        plt.show()

    # Plot performance over actual time
    for gid, gdf in perf_time.groupby(by=["scenario", "set"]):
        print(gid)
        fig, ax = plot_performance_over_time(
            df=gdf,
            x="time_norm",
            y="trial_value__cost_inc_norm",
            hue="optimizer_id",
            figure_filename=f"figures/perf_over_time/performance_over_time_{gid}_elapsed.pdf",
            figsize=(6,4),
            **lineplot_kwargs
        )